#include <stdio.h>
#include <hash_map.h>
#include <stdlib.h>
#include <multimap.h>
#include <string>

#define BUF_SIZE 4096

struct eqstr
{
  bool operator()(const char *s1, const char *s2) const
  { return strcmp(s1, s2) == 0; }
};

hash_map<const char *, const char *, hash<const char *>, eqstr> months;

class LogFile
{
public:
  LogFile();
  ~LogFile(){ if(m_fp) fclose(m_fp); }

  // open a file, return 0 for success, 1 for minor error (0 length file)
  // and 2 for serious error (file not found)
  int open(const char *name);

  // get the date string of the current item
  const char *date() const { return m_date; }
  // get the full data for the current item
  const char *line() const { return m_line; }

  // get a line from the file and parse it
  // return 0 for success and 1 for EOF
  int getLine();

private:
  FILE *m_fp;
  char m_date[17];
  char m_line[BUF_SIZE];

  // store the date in numeric format so that strcmp() can be used to compare
  // dates
  int setDate();

  LogFile(const LogFile&);
  LogFile & operator=(const LogFile&);
};

LogFile::LogFile()
 : m_fp(NULL)
{
  m_line[0] = '\0';
}

// Common log format:
// 1.2.3.4 - - [23/Aug/2000:12:00:32 +0200] etc
int LogFile::setDate()
{
  unsigned int i;
  // find '[' or return if we can't
  for(i = 0; m_line[i] != '['; i++)
    if(m_line[i] == '\0') return 1;
  // if not enough data left for the full date then return
  if(i + 21 > strlen(m_line))
    return 1;
  memcpy(m_date, &m_line[i + 8], 4);
  char mon[4];
  memcpy(mon, &m_line[i + 4], 3);
  mon[3] = '\0';
  const char *m = months[mon];
  if(!m) return 1;
  strcpy(&m_date[4], m);
  memcpy(&m_date[6], &m_line[i + 1], 2);
  memcpy(&m_date[8], &m_line[i + 13], 8);
  m_date[16] = '\0';
  return 0;
}

int LogFile::open(const char *name)
{
  m_fp = fopen(name, "r");
  if(!m_fp)
  {
    fprintf(stderr, "Can't open %s.\n", name);
    return 2;
  }
  if(getLine())
    return 1;
  return 0;
}

int LogFile::getLine()
{
  while(1)
  {
    // if can't get more data then return 1
    if(!fgets(m_line, sizeof(m_line) - 1, m_fp))
      return 1;
    m_line[sizeof(m_line) - 1] = '\0';
    strtok(m_line, "\n");
    // if setDate() returns 1 then we can't parse the line so we keep looping
    if(!setDate())
      return 0;
    fprintf(stderr, "Skipping bad line: %s\n", m_line);
  }
  return 0;
}

typedef LogFile *PLogFile;

int item_compare(const void *a, const void *b)
{
  const LogFile * const left = *(LogFile * const *)a;
  const LogFile * const right = *(LogFile * const *)b;
  return strcmp(left->date(), right->date());
}

struct ltstr
{
  bool operator()(const string s1, const string s2) const
  {
    return strcmp(s1.c_str(), s2.c_str()) < 0;
  }
};

int main(int argc, char **argv)
{
  if(argc == 1)
    return 0;
  if(argc == 2 && (!strcmp(argv[1], "-h") || !strcmp(argv[1], "--help")) )
  {
    fprintf(stderr, "usage: %s [filenames]", argv[0]);
    fprintf(stderr, "\n"
    "This program merges web logs in common log format into a single stream\n"
    "on standard output.  It reads from multiple input files and outputs the\n"
    "data in-order as much as is possible.  If there is only a single input\n"
    "file it will re-order it (with a 1000 line buffer size) to deal with web\n"
    "servers that output data out of order.\n");
    return 0;
  }
  months["Jan"] = "01";
  months["Feb"] = "02";
  months["Mar"] = "03";
  months["Apr"] = "04";
  months["May"] = "05";
  months["Jun"] = "06";
  months["Jul"] = "07";
  months["Aug"] = "08";
  months["Sep"] = "09";
  months["Oct"] = "10";
  months["Nov"] = "11";
  months["Dec"] = "12";

  multimap<const string, const string, ltstr> outputMap;

  LogFile **items = new PLogFile[argc - 1];

  unsigned int item_count = 0;
  int i;
  for(i = 1; i < argc; i++)
  {
    items[item_count] = new LogFile;
    int rc = items[item_count]->open(argv[i]);
    if(rc > 1)
      return rc;
    if(rc == 1)
      delete items[item_count];
    else
      item_count++;
  }

  unsigned int map_items = item_count * 400;
  if(map_items < 4000)
    map_items = 4000;
  while(item_count > 1)
  {
    qsort(items, item_count, sizeof(LogFile *), item_compare);
    while(strcmp(items[0]->date(), items[1]->date()) <= 0)
    {
      outputMap.insert(pair<string, string>(items[0]->date(), items[0]->line()));
      while(outputMap.size() > map_items)
      {
        printf("%s\n", outputMap.begin()->second.c_str());
        outputMap.erase(outputMap.begin());
      }
      if(items[0]->getLine())
      {
        delete(items[0]);
        item_count--;
        items[0] = items[item_count];
        break;
      }
    }
  }
  if(item_count == 1)
  {
    do
    {
      outputMap.insert(pair<string, string>(items[0]->date(), items[0]->line()));
    } while(!items[0]->getLine() && outputMap.size() < map_items);

    do
    {
      outputMap.insert(pair<string, string>(items[0]->date(), items[0]->line()));
      printf("%s\n", outputMap.begin()->second.c_str());
      outputMap.erase(outputMap.begin());
    } while(!items[0]->getLine());
    delete items[0];
    while(!outputMap.empty())
    {
      printf("%s\n", outputMap.begin()->second.c_str());
      outputMap.erase(outputMap.begin());
    }
  }
  delete items;
  return 0;
}
